import configparser
import logging
import os.path

import boto3
import botocore
import botocore.config
import requests

from .auth import login_aws_via_idp

logger = logging.getLogger(__name__)


def write_aws_config(profile_name, region, executable):
    config_filepath = os.path.expanduser("~/.aws/config")
    config = configparser.RawConfigParser()
    config.read(config_filepath)

    section_name = f"profile {profile_name}"
    if not config.has_section(section_name):
        config.add_section(section_name)

    config.set(section_name, 'region', region)

    cred_process = f"{executable} login"
    config.set(section_name, "credential_process", cred_process)

    # Write the updated config file.
    with open(config_filepath, 'w+') as configfile:
        config.write(configfile)


def get_aws_creds(username, password, entity_id, role_arn):
    assertion, roles = login_aws_via_idp(
        requests.Session(), username, password, entity_id)
    arns = parse_roles(roles)
    if role_arn not in arns:
        logger.debug("all arns: %s", arns)
        raise ValueError(f"you do not have permission to assume {role_arn}")
    principal_arn = arns[role_arn]
    return sts_creds_from_saml(role_arn, principal_arn, assertion)


def parse_roles(roles):
    arns = {}
    for r in roles:
        role_arn, principal_arn = r.split(",", 1)
        arns[role_arn] = principal_arn
    return arns


def sts_creds_from_saml(role_arn, principal_arn, saml_assertion):
    # Skip using any credentials on disk.
    client_conf = botocore.config.Config(
        signature_version=botocore.UNSIGNED,
    )
    client = boto3.client("sts", config=client_conf)
    token = client.assume_role_with_saml(
        RoleArn=role_arn,
        PrincipalArn=principal_arn,
        SAMLAssertion=saml_assertion,
    )
    return token["Credentials"]



import click
import os.path
import json
import logging

from .awscreds import get_aws_creds
from .configs import Config, default_config_file, default_cache_dir
from .credentials import CredentialSet


logger = logging.getLogger(__name__)


@click.command()
@click.option("--config-file",
              default=default_config_file(),
              help="filename to store configuration information",
              type=click.Path(file_okay=True, writable=True),
              show_default=True)
@click.option("--cache-ignore", is_flag=True,
              help="ignore any cached credentials, and dont store in cache")
@click.option("--cache-force-refresh", is_flag=True,
              help="force an update of cached credentials")
@click.option("--cache-dir", default=default_cache_dir(),
              help="directory to cache credentials in")
def login(config_file, cache_ignore, cache_force_refresh, cache_dir):
    if not os.path.exists(config_file):
        raise click.UsageError("Config not setup. Run scimma-aws setup first.")
    config = Config.from_file(config_file)

    creds = None
    creds_cache_filename = cache_dir / config.profile_name
    if not cache_ignore and not cache_force_refresh:
        try:
            logger.debug("checking for credentials in cache at %s",
                         creds_cache_filename)
            creds = CredentialSet.from_cache_file(creds_cache_filename)
        except FileNotFoundError:
            logger.debug("cached credentials not found")
            pass

    if creds is not None:
        # Cached credentials found. Check if they're still live.
        logger.debug("cached credentials found, checking expiration")
        if creds.expired():
            # No good.
            logger.debug("credentials are expired")
            creds = None

    if creds is None:
        logger.debug("refreshing credentials")
        raw_creds = get_aws_creds(
            config.username, config.password, config.entity_id, config.role_arn)
        creds = CredentialSet.from_aws_creds(raw_creds)
        if not cache_ignore:
            logger.debug("storing credentials in cache")
            creds.to_cache_file(creds_cache_filename)

    print(json.dumps(creds.to_aws_creds()))



import click
import os.path
import requests
import logging
import sys

from typing import Dict

from .auth import login_aws_via_idp
from .awscreds import parse_roles, write_aws_config
from .configs import Config, default_config_file


logger = logging.getLogger(__name__)


@click.command()
@click.option("--config",
              default=default_config_file(),
              help="filename to store configuration information",
              type=click.Path(file_okay=True, writable=True),
              show_default=True)
def setup(config):
    if os.path.exists(config):
        confirmed = click.confirm(
            "Config file already exists. Proceeding will overwrite. OK?")
        if not confirmed:
            return

    # Prompt for IdP Entity ID
    click.echo(click.wrap_text(
        "You'll need to provide your SAML IdP entity ID. Details on campus "
        "IdPs including the SAML entity ID may be found at "
        "https://incommon.org/federation/incommon-federation-entities/.",
    ),
    )
    click.echo()
    print_well_known_idps()
    entity_id = click.prompt("Enter your IdP Entity ID", type=str).strip()

    # Prompt for username and password
    username = click.prompt(
        "Username you use to login to your IdP", type=str).strip()
    password = click.prompt("Password", hide_input=True, type=str)

    region = click.prompt("AWS Region to connect to",
                          type=str, default="us-west-2").strip()
    profile_name = click.prompt(
        "AWS credential profile name to use", type=str, default="default")
    profile_name = profile_name.strip()

    conf = Config(
        entity_id=entity_id,
        username=username,
        password=password,
        region=region,
        profile_name=profile_name,
    )

    click.secho(
        "Verifying your credentials by attempting to log in.", color="yellow")
    assertion, roles = login_aws_via_idp(
        requests.Session(), conf.username, conf.password, conf.entity_id,
    )
    click.secho("Log-in succesful!", color="green")

    roles = parse_roles(roles)
    if len(roles) == 1:
        conf.role_arn = list(roles.keys())[0]
    else:
        conf.role_arn = prompt_for_role(roles)
    click.echo(f"You will log in as {conf.role_arn}.")

    conf.to_file(config)
    click.echo(f"Your configuration has been saved to {config}.")
    write_aws_config(conf.profile_name, conf.region, sys.argv[0])
    click.echo("AWS configuration has been saved.")


def prompt_for_role(roles: Dict[str, str]):
    """ Prompt the user to pick one of the provided roles. """
    click.echo("Choose role to assume")
    enumerated = dict(enumerate(roles.keys(), start=1))
    for i, role_arn in sorted(enumerated.items()):
        click.echo(f"  [{i}]: {role_arn}")
        chosen_idx = click.prompt(
            "Choose role to assume", type=int,
        )
    return enumerated[chosen_idx]


def print_well_known_idps():
    click.echo("Some well-known IdP entity IDs:")
    well_known_idps = {
        "Cornell University": "https://shibidp.cit.cornell.edu/idp/shibboleth",
        "LIGO": "https://login.ligo.org/idp/shibboleth",
        "Penn State": "urn:mace:incommon:psu.edu",
        "University of California, Santa Barbara": "urn:mace:incommon:ucsb.edu",
        "University of Illinois at Urbana-Champaign": "urn:mace:incommon:uiuc.edu",
        "University of Washington": "urn:mace:incommon:washington.edu",
        "University of Wisconsin-Milwaukee": "https://idp.uwm.edu/idp/shibboleth",
    }
    for institution, entity in well_known_idps.items():
        click.secho(f"  {institution}", fg="yellow", nl=False)
        click.secho(f": {entity}")



import dataclasses
import datetime
import os
import json


@dataclasses.dataclass
class CredentialSet:
    version: int
    access_key_id: str
    secret_access_key: str
    session_token: str
    expiration: datetime.datetime

    def to_aws_creds(self):
        return {
            "Version": self.version,
            "AccessKeyId": self.access_key_id,
            "SecretAccessKey": self.secret_access_key,
            "SessionToken": self.session_token,
            "Expiration": self.expiration.isoformat()
        }

    def to_cache_file(self, filename):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        json_data = self.to_aws_creds()
        with open(filename, "w") as f:
            json.dump(json_data, f)

    @classmethod
    def from_aws_creds(cls, creds):
        try:
            expiration = datetime.datetime.fromisoformat(creds["Expiration"])
        except TypeError:
            expiration = creds["Expiration"]
        return CredentialSet(
            version=1,
            access_key_id=creds["AccessKeyId"],
            secret_access_key=creds["SecretAccessKey"],
            session_token=creds["SessionToken"],
            expiration=expiration,
        )

    @classmethod
    def from_cache_file(cls, filename):
        with open(filename, "r") as f:
            json_data = json.load(f)
            return CredentialSet.from_aws_creds(json_data)

    def expired(self, margin=datetime.timedelta(minutes=10)):
        return datetime.datetime.now(self.expiration.tzinfo) > (self.expiration - margin)



from base64 import b64encode, b64decode
from bs4 import BeautifulSoup
from datetime import datetime
from random import randrange
import re
import logging
import xml.etree.ElementTree as ET


SCIMMA_PROXY_SSO = "https://federation-proxy.scimma.org/cilogon/sso/post"
logger = logging.getLogger(__name__)


def login_aws_via_idp(session, username, password, entity_id):
    """ Get a SAML assertion and set of AWS roles which can be assumed with the SAML assertion. """
    logger.info("Looking up your IdP")
    idp_url, idp_form = get_idp_login_form(session, username, password, entity_id)

    logger.info("Logging in to %s", idp_url)
    idp_response = session.post(idp_url, data=idp_form)
    idp_response.raise_for_status()

    logger.info("Parsing response and presenting assertion to CILogon")
    cilogon_url, payload = parse_idp_login_response(idp_response.text)
    scimma_saml_proxy_response = session.post(cilogon_url, data=payload)
    scimma_saml_proxy_response.raise_for_status()

    logger.info("Login complete, extracting credentials")
    assertion = parse_scimma_sample_response(scimma_saml_proxy_response.text)
    roles = parse_scimma_aws_assertion(assertion)
    return assertion, roles


def get_idp_login_form(session, username, password, entity_id):
    url, payload = get_idp_discovery_form(session)
    payload["providerId"] = entity_id
    idp_form_response = session.post(url, data=payload, verify=True)
    idp_form_response.raise_for_status()
    idp_form = populate_idp_form(idp_form_response.text, username, password)
    return (idp_form_response.url, idp_form)


def get_idp_discovery_form(session):
    """Make a SAML auth request, follow the redirects, and return a CILogon IdP
    discovery form along with the URL to send it to.

    Uses the provided requests Session, which should follow redirects.
    """
    cilogon_form_response = post_authn_request(session)
    cilogon_form_response.raise_for_status()
    form_contents = parse_lxml_form(cilogon_form_response.text)
    return (cilogon_form_response.url, form_contents)


def parse_lxml_form(text):
    """ Parse lxml form into a dictionary of name:value based on the 'input' tags in text"""
    soup = BeautifulSoup(text, features="lxml")
    form = {}
    for inputtag in soup.find_all(re.compile("(INPUT|input)")):
        name = inputtag.get("name", "")
        value = inputtag.get("value", "")
        form[name] = value
    return form


def post_authn_request(session):
    """ Post a SAML authentication request to the SCIMMA SAML Proxy IdP.

    Uses the provided requests Session, which should follow redirects.
    """
    req = build_authn_request()
    return session.post(SCIMMA_PROXY_SSO, data={"SAMLRequest": req}, verify=True)


def build_authn_request():
    message_id = "_%030x" % randrange(16**30)
    timestamp = datetime.now().replace(microsecond=0).isoformat() + "Z"
    authn_request = f"""
<samlp:AuthnRequest
  xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
  xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
  AssertionConsumerServiceURL="https://signin.aws.amazon.com/saml"
  Destination="{SCIMMA_PROXY_SSO}"
  ID="{message_id}"
  IssueInstant="{timestamp}"
  ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
  Version="2.0">
  <saml:Issuer>urn:amazon:webservices</saml:Issuer>
    <samlp:NameIDPolicy AllowCreate="1" />
</samlp:AuthnRequest>

    """.encode()
    authn_base64 = b64encode(authn_request).decode("utf-8")
    return authn_base64


def populate_idp_form(form_text, username, password):
    """Try to populate an lxml login form. Make educated guesses on the username
    and password field names."""
    form = parse_lxml_form(form_text)
    result = {}
    for key, value in form.items():
        if "user" in key.lower():
            result[key] = username
        elif "email" in key.lower():
            result[key] = username
        elif "pass" in key.lower():
            result[key] = password
        else:
            result[key] = value
    # Shibboleth IdPs will need to include this field.
    result["_eventId_proceed"] = ""

    return result


def parse_idp_login_response(text):
    # Decode the response from the authenticating IdP and extract the SAML
    # assertion
    soup = BeautifulSoup(text, features="lxml")
    assertion = ''

    for inputtag in soup.find_all('input'):
        if (inputtag.get('name') == 'SAMLResponse'):
            assertion = inputtag.get('value')

        if (inputtag.get('name') == 'RelayState'):
            relay_state = inputtag.get('value')
    if assertion == "":
        raise ValueError("Response did not contain a valid SAML assertion")

    # Parse the returned payload to find the AssertionConsumerService URL
    # for the CILogon SP to which we will POST the SAML Response. Since CILogon
    # uses multiple SP ACS URLs we do not hardcode this.
    cilogon_sp_acs_url = soup.find('form').get('action')
    payload = {
        "SAMLResponse": assertion,
        "RelayState": relay_state,
    }
    return (cilogon_sp_acs_url, payload)


def parse_scimma_sample_response(text):
    """Extract the SAML Response assertion from from a CILogon SAML Proxy response. """
    soup = BeautifulSoup(text, features="lxml")
    assertion = ''

    for inputtag in soup.find_all('input'):
        if (inputtag.get('name') == 'SAMLResponse'):
            assertion = inputtag.get('value')
    return assertion


def parse_scimma_aws_assertion(assertion):
    """ Decode a SAML assertion and extract the AWS Roles it grants."""
    attr_element = '{urn:oasis:names:tc:SAML:2.0:assertion}Attribute'
    attr_value_element = '{urn:oasis:names:tc:SAML:2.0:assertion}AttributeValue'
    role_attribute_name = 'https://aws.amazon.com/SAML/Attributes/Role'
    aws_roles = []
    root = ET.fromstring(b64decode(assertion))
    for saml2attribute in root.iter(attr_element):
        if (saml2attribute.get('Name') == role_attribute_name):
            for saml2attributevalue in saml2attribute.iter(attr_value_element):
                aws_roles.append(saml2attributevalue.text)
    return aws_roles



import logging

import click

from .cmd_setup import setup
from .cmd_login import login

logger = logging.getLogger(__name__)


@click.group()
@click.option("--verbose", is_flag=True)
@click.option("--debug", is_flag=True)
def main(verbose, debug):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose:
        logging.basicConfig(level=logging.INFO)
    pass


main.add_command(setup)
main.add_command(login)



import dataclasses
import json
import os
import os.path
import pathlib


def default_config_dir() -> pathlib.Path:
    if os.getenv("XDG_CONFIG_HOME"):
        config_root = pathlib.Path(os.getenv("XDG_CONFIG_HOME"))
    else:
        config_root = pathlib.Path.home() / ".config"
    return config_root / "scimma-aws"


def default_config_file() -> pathlib.Path:
    return default_config_dir() / "config"


def default_cache_dir() -> pathlib.Path:
    return default_config_dir() / "cache"


@dataclasses.dataclass
class Config:
    username: str
    password: str
    entity_id: str
    region: str
    profile_name: str
    role_arn: str = ""

    def to_file(self, filepath):
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        config_data = {
            "username": self.username,
            "password": self.password,
            "entity_id": self.entity_id,
            "region": self.region,
            "profile_name": self.profile_name,
            "role_arn": self.role_arn,
        }
        with open(filepath, "w") as f:
            json.dump(config_data, f)
        os.chmod(filepath, 0o600)

    @classmethod
    def from_file(cls, filepath):
        data = json.load(open(filepath, "r"))
        return Config(**data)
